@@ -1331,6 +1331,58 @@ void CalendarView::toggleTodoCompleted(const Akonadi::Item &todoItem)
     mChanger->endAtomicOperation();
 }
 
+void CalendarView::toggleCompleted(KCalendarCore::Todo::Ptr todo, const QDate &occurrenceDate)
+{
+    if (todo->recurs()) {
+        QDateTime recurrenceId = recurrenceOnDate(todo, occurrenceDate);
+        const auto dtRecurrence = todo->dtRecurrence();
+        if (todo->isCompleted()) {
+            todo->setCompleted(false);
+            todo->setDtRecurrence(recurrenceId);
+        } else if (dtRecurrence <= recurrenceId) {
+            // Occurrence is not complete.
+            const auto dtNextOccurrence = todo->recurrence()->getNextDateTime(recurrenceId);
+            if (dtNextOccurrence.isValid()) {
+                todo->setDtRecurrence(dtNextOccurrence);
+            } else {
+                todo->setDtRecurrence(recurrenceId);
+                todo->setCompleted(true);
+            }
+        } else {
+            todo->setDtRecurrence(recurrenceId);
+        }
+    } else {
+        if (todo->isCompleted()) {
+            todo->setPercentComplete(0);
+            todo->setCompleted(false);
+            todo->setStatus(KCalendarCore::Incidence::StatusNone);
+        } else {
+            todo->setPercentComplete(0);
+            todo->setCompleted(QDateTime::currentDateTime());
+            todo->setStatus(KCalendarCore::Incidence::StatusCompleted);
+        }
+    }
+}
+
+void CalendarView::toggleOccurrenceCompleted(const Akonadi::Item &todoItem, const QDate &occurrenceDate)
+{
+    KCalendarCore::Todo::Ptr todo = Akonadi::CalendarUtils::todo(todoItem);
+    if (!todo) {
+        qWarning(KORGANIZER_LOG) << "item does not contain a todo.";
+        return;
+    }
+    if (todo->recurs() && !occurrenceDate.isValid()) {
+        qWarning(KORGANIZER_LOG) << "todo recurs, but no occurrence date was provided.";
+        return;
+    }
+
+    KCalendarCore::Todo::Ptr oldtodo {todo->clone()};
+    toggleCompleted(todo, occurrenceDate);
+    mChanger->startAtomicOperation(i18n("Toggle To-do Completed"));
+    (void)mChanger->modifyIncidence(todoItem, oldtodo, this);
+    mChanger->endAtomicOperation();
+}
+
 void CalendarView::copyIncidenceToResource(const Akonadi::Item &item, const Akonadi::Collection &col)
 {
 #ifdef AKONADI_PORT_DISABLED
@@ -1473,6 +1525,13 @@ void CalendarView::moveIncidenceToResource(const Akonadi::Item &item, const Akon
 #endif
 }
 
+QDateTime CalendarView::recurrenceOnDate(KCalendarCore::Incidence::Ptr incidence, QDate displayDate)
+{
+    const auto start = incidence->dateTime(KCalendarCore::IncidenceBase::RoleDisplayStart);
+    const auto offset = start.toLocalTime().date().daysTo(displayDate);
+    return incidence->dateTime(KCalendarCore::IncidenceBase::RoleRecurrenceStart).addDays(offset);
+}
+
 void CalendarView::dissociateOccurrences(const Akonadi::Item &item, QDate date)
 {
     const KCalendarCore::Incidence::Ptr incidence = Akonadi::CalendarUtils::incidence(item);
@@ -1482,8 +1541,8 @@ void CalendarView::dissociateOccurrences(const Akonadi::Item &item, QDate date)
         return;
     }
 
-    QDateTime thisDateTime(date, {}, Qt::LocalTime);
-    bool isFirstOccurrence = !incidence->recurrence()->getPreviousDateTime(thisDateTime).isValid();
+    QDateTime recurrenceId = recurrenceOnDate(incidence, date);
+    bool isFirstOccurrence = !incidence->recurrence()->getPreviousDateTime(recurrenceId).isValid();
 
     int answer;
     bool doOnlyThis = false;
@@ -1516,13 +1575,13 @@ void CalendarView::dissociateOccurrences(const Akonadi::Item &item, QDate date)
     }
 
     if (doOnlyThis) {
-        dissociateOccurrence(item, date, false);
+        dissociateOccurrence(item, recurrenceId, false);
     } else if (doFuture) {
-        dissociateOccurrence(item, date, true);
+        dissociateOccurrence(item, recurrenceId, true);
     }
 }
 
-void CalendarView::dissociateOccurrence(const Akonadi::Item &item, const QDate &date, bool thisAndFuture)
+void CalendarView::dissociateOccurrence(const Akonadi::Item &item, const QDateTime &recurrenceId, bool thisAndFuture)
 {
     const KCalendarCore::Incidence::Ptr incidence = Akonadi::CalendarUtils::incidence(item);
 
@@ -1531,10 +1590,7 @@ void CalendarView::dissociateOccurrence(const Akonadi::Item &item, const QDate &
     } else {
         startMultiModify(i18n("Dissociate occurrence"));
     }
-    QDateTime occurrenceDate = incidence->dtStart();
-    occurrenceDate.setDate(date);
-    qCDebug(KORGANIZER_LOG) << "create exception: " << occurrenceDate;
-    KCalendarCore::Incidence::Ptr newInc(KCalendarCore::Calendar::createException(incidence, occurrenceDate, thisAndFuture));
+    KCalendarCore::Incidence::Ptr newInc(KCalendarCore::Calendar::createException(incidence, recurrenceId, thisAndFuture));
     if (newInc) {
         (void)mChanger->createIncidence(newInc, item.parentCollection(), this);
     } else {
